import time
import pytest
import unittest
import pandas as pd
from pathlib import Path
import os
from indico.client import IndicoClient
from indico.queries.datasets import (
    GetDataset,
    GetDatasetFileStatus,
    CreateDataset,
    ListDatasets,
    DeleteDataset,
    CreateEmptyDataset,
    AddFiles,
    ProcessFiles,
    ProcessCSV,
)
from indico.queries.export import CreateExport, DownloadExport
from indico.types.dataset import Dataset
from indico.errors import IndicoRequestError
from tests.integration.data.datasets import airlines_dataset


def test_create_dataset(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[1]) + "/data/AirlineComplaints.csv"

    response = client.call(
        CreateDataset(
            name=f"AirlineComplaints-test-{int(time.time())}", files=[dataset_filepath]
        )
    )

    assert isinstance(response, Dataset)
    assert response.status == "COMPLETE"
    assert isinstance(response.id, int)


def test_get_datasets(indico, airlines_dataset):
    client = IndicoClient()
    dataset = client.call(GetDataset(id=airlines_dataset.id))

    assert isinstance(dataset, Dataset)
    assert dataset.id == airlines_dataset.id


def test_get_dataset_with_bad_id(indico):
    client = IndicoClient()

    with pytest.raises(IndicoRequestError):
        dataset = client.call(GetDataset(id=10000000))


def test_get_dataset_file_status(indico, airlines_dataset):
    client = IndicoClient()
    dataset = client.call(GetDatasetFileStatus(id=airlines_dataset.id))

    assert isinstance(dataset, Dataset)
    assert dataset.id == airlines_dataset.id
    assert len(dataset.files) > 0
    assert dataset.files[0].status != None


def test_list_datasets(indico, airlines_dataset):
    client = IndicoClient()
    datasets = client.call(ListDatasets(limit=1))

    assert isinstance(datasets, list)
    assert len(datasets) == 1
    assert type(datasets[0]) == Dataset


def test_images(indico):
    client = IndicoClient()

    dataset_filepath = str(Path(__file__).parents[1]) + "/data/dog_vs_cats_small.csv"
    response = client.call(
        CreateDataset(
            name=f"image-dataset-test-{int(time.time())}",
            files=dataset_filepath,
            from_local_images=True,
        )
    )
    assert isinstance(response, Dataset)
    assert response.status == "COMPLETE"
    assert isinstance(response.id, int)


def test_images_batch(indico):
    client = IndicoClient()

    dataset_filepath = str(Path(__file__).parents[1]) + "/data/dog_vs_cats_small.csv"
    response = client.call(
        CreateDataset(
            name=f"image-dataset-test-{int(time.time())}",
            files=dataset_filepath,
            from_local_images=True,
            batch_size=10,
        )
    )
    assert isinstance(response, Dataset)
    assert response.status == "COMPLETE"
    assert isinstance(response.id, int)


def test_upload_pdf_dataset_batch(indico):
    client = IndicoClient()
    file_names = ["mock.pdf", "mock_2.pdf", "mock_3.pdf"]
    parent_path = str(Path(__file__).parent.parent / "data")
    dataset_filepaths = [
        os.path.join(parent_path, file_name) for file_name in file_names
    ]
    dataset = client.call(
        CreateDataset(
            name=f"pdf-dataset-test-{int(time.time())}",
            files=dataset_filepaths,
            batch_size=1,
        )
    )
    assert isinstance(dataset, Dataset)
    assert dataset.status == "COMPLETE"
    assert isinstance(dataset.id, int)
    assert dataset.datacolumns[0].name == "text"
    response = client.call(GetDatasetFileStatus(id=dataset.id))
    for datafile in response.files:
        assert datafile.status == "PROCESSED"
        assert datafile.file_type == "PDF"


def test_upload_pdf_interrupt(indico):
    client = IndicoClient()

    client._http.request_session.cookies.clear()
    file_names = ["mock.pdf", "mock_2.pdf", "mock_3.pdf"]
    parent_path = str(Path(__file__).parent.parent / "data")
    dataset_filepaths = [
        os.path.join(parent_path, file_name) for file_name in file_names
    ]

    dataset = client.call(
        CreateDataset(
            name=f"pdf-dataset-test-{int(time.time())}",
            files=dataset_filepaths,
            batch_size=1,
        )
    )


def test_delete_dataset(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[1]) + "/data/AirlineComplaints.csv"

    dataset = client.call(
        CreateDataset(
            name=f"AirlineComplaints-test-{int(time.time())}", files=[dataset_filepath]
        )
    )

    assert isinstance(dataset, Dataset)
    assert dataset.status == "COMPLETE"
    assert isinstance(dataset.id, int)

    success = client.call(DeleteDataset(id=dataset.id))
    assert success == True

    with pytest.raises(IndicoRequestError):
        dataset = client.call(GetDataset(id=dataset.id))


def _dataset_complete(dataset):
    for df in dataset.files:
        assert df.status == "PROCESSED"

    assert dataset.status == "COMPLETE"


def test_create_from_files_document(indico):
    client = IndicoClient()
    dataset = client.call(CreateEmptyDataset(name=f"dataset-{int(time.time())}"))
    file_names = ["us_doi.tiff", "mock.pdf"]
    parent_path = str(Path(__file__).parent.parent / "data")
    dataset_filepaths = [
        os.path.join(parent_path, file_name) for file_name in file_names
    ]

    dataset = client.call(AddFiles(dataset_id=dataset.id, files=[dataset_filepaths[0]]))

    for f in dataset.files:
        assert f.status == "DOWNLOADED"

    datafile_ids = [f.id for f in dataset.files]

    dataset = client.call(
        ProcessFiles(dataset_id=dataset.id, datafile_ids=datafile_ids, wait=True)
    )

    dataset = client.call(AddFiles(dataset_id=dataset.id, files=[dataset_filepaths[1]]))

    datafile_ids = [f.id for f in dataset.files if f.status == "DOWNLOADED"]

    dataset = client.call(
        ProcessFiles(dataset_id=dataset.id, datafile_ids=datafile_ids, wait=True)
    )

    _dataset_complete(dataset)


def test_create_from_files_image(indico):
    client = IndicoClient()
    dataset = client.call(
        CreateEmptyDataset(name=f"dataset-{int(time.time())}", dataset_type="IMAGE")
    )
    file_names = ["1.jpg", "2.jpg", "3.jpg"]
    parent_path = str(Path(__file__).parent.parent / "data")
    dataset_filepaths = [
        os.path.join(parent_path, file_name) for file_name in file_names
    ]

    dataset = client.call(AddFiles(dataset_id=dataset.id, files=dataset_filepaths))

    for f in dataset.files:
        assert f.status == "DOWNLOADED"

    datafile_ids = [f.id for f in dataset.files]

    dataset = client.call(
        ProcessFiles(dataset_id=dataset.id, datafile_ids=datafile_ids, wait=True)
    )

    file_names = ["4.jpg", "5.jpg"]
    dataset_filepaths = [
        os.path.join(parent_path, file_name) for file_name in file_names
    ]

    dataset = client.call(AddFiles(dataset_id=dataset.id, files=dataset_filepaths))

    datafile_ids = [f.id for f in dataset.files if f.status == "DOWNLOADED"]

    dataset = client.call(
        ProcessFiles(dataset_id=dataset.id, datafile_ids=datafile_ids, wait=True)
    )

    _dataset_complete(dataset)


def test_create_from_csv(indico):
    client = IndicoClient()
    dataset = client.call(CreateEmptyDataset(name=f"dataset-{int(time.time())}"))
    dataset_filepath = str(Path(__file__).parents[1]) + "/data/AirlineComplaints.csv"
    dataset = client.call(AddFiles(dataset_id=dataset.id, files=[dataset_filepath]))

    for f in dataset.files:
        assert f.status == "DOWNLOADED"

    datafile_ids = [f.id for f in dataset.files]

    dataset = client.call(
        ProcessCSV(dataset_id=dataset.id, datafile_ids=datafile_ids, wait=True)
    )

    dataset_filepath = str(Path(__file__).parents[1]) + "/data/AirlineComplaints.csv"
    dataset = client.call(AddFiles(dataset_id=dataset.id, files=[dataset_filepath]))
    datafile_ids = [f.id for f in dataset.files if f.status == "DOWNLOADED"]

    dataset = client.call(
        ProcessCSV(dataset_id=dataset.id, datafile_ids=datafile_ids, wait=True)
    )

    _dataset_complete(dataset)

    export = client.call(CreateExport(dataset_id=dataset.id, wait=True))

    exported_data = client.call(DownloadExport(export.id))

    baseline_data = pd.read_csv(dataset_filepath)

    for col in baseline_data.columns:
        assert all(
            baseline_data[col].apply(str).str.strip().values
            == exported_data[: len(baseline_data)][col].apply(str).values
        )
        assert all(
            baseline_data[col].apply(str).str.strip().values
            == exported_data[len(baseline_data) :][col].apply(str).values
        )


def test_create_from_csv_image_urls(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[1]) + "/data/image_link_small.csv"
    dataset = client.call(
        CreateEmptyDataset(name=f"dataset-{int(time.time())}", dataset_type="IMAGE")
    )
    dataset = client.call(AddFiles(dataset_id=dataset.id, files=[dataset_filepath]))

    for f in dataset.files:
        assert f.status == "DOWNLOADED"

    datafile_ids = [f.id for f in dataset.files]

    dataset = client.call(
        ProcessCSV(dataset_id=dataset.id, datafile_ids=datafile_ids, wait=True)
    )

    _dataset_complete(dataset)


def test_create_from_csv_image_urls_with_broken(indico):
    client = IndicoClient()
    dataset_filepath = (
        str(Path(__file__).parents[1]) + "/data/image_link_small_with_broken.csv"
    )
    dataset = client.call(
        CreateEmptyDataset(name=f"dataset-{int(time.time())}", dataset_type="IMAGE")
    )
    dataset = client.call(AddFiles(dataset_id=dataset.id, files=[dataset_filepath]))

    for f in dataset.files:
        assert f.status == "DOWNLOADED"

    datafile_ids = [f.id for f in dataset.files]

    dataset = client.call(
        ProcessCSV(dataset_id=dataset.id, datafile_ids=datafile_ids, wait=True)
    )

    failed = 0
    for df in dataset.files:
        if df.status == "FAILED":
            assert df.failure_type == "DOWNLOAD"
            failed += 1
        else:
            assert df.status == "PROCESSED"
    assert failed == 1


def test_create_from_csv_doc_urls(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[1]) + "/data/pdf_links.csv"
    dataset = client.call(CreateEmptyDataset(name=f"dataset-{int(time.time())}"))
    dataset = client.call(AddFiles(dataset_id=dataset.id, files=[dataset_filepath]))

    for f in dataset.files:
        assert f.status == "DOWNLOADED"

    datafile_ids = [f.id for f in dataset.files]

    dataset = client.call(
        ProcessCSV(dataset_id=dataset.id, datafile_ids=datafile_ids, wait=True)
    )

    _dataset_complete(dataset)


@unittest.skip
def test_csv_incompat_columns(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[1]) + "/data/pdf_links.csv"
    dataset = client.call(CreateEmptyDataset(name=f"dataset-{int(time.time())}"))
    dataset = client.call(AddFiles(dataset_id=dataset.id, files=[dataset_filepath]))

    for f in dataset.files:
        assert f.status == "DOWNLOADED"

    datafile_ids = [f.id for f in dataset.files]

    dataset = client.call(
        ProcessCSV(dataset_id=dataset.id, datafile_ids=datafile_ids, wait=True)
    )

    dataset_filepath = str(Path(__file__).parents[1]) + "/data/AirlineComplaints.csv"
    dataset = client.call(AddFiles(dataset_id=dataset.id, files=[dataset_filepath]))

    datafile_ids = [
        df.id
        for df in dataset.files
        if df.status == "DOWNLOADED" and df.file_type == "CSV"
    ]

    dataset = client.call(
        ProcessCSV(dataset_id=dataset.id, datafile_ids=datafile_ids, wait=True)
    )

    failed = 0
    for df in dataset.files:
        if df.status == "FAILED":
            assert df.failure_type == "INCOMPATIBLE_CSV_COLUMNS"
            failed += 1
        else:
            assert df.status == "PROCESSED"

    assert dataset.status == "COMPLETE"


def test_bad_csv_create_dataset(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[1]) + "/data/classification.csv"
    dataset = client.call(
        CreateDataset(
            name=f"dataset-{int(time.time())}", files=[dataset_filepath], wait=True
        ),
    )

    assert dataset.status == "CREATING"
    dataset = client.call(GetDatasetFileStatus(id=dataset.id))
    assert all([f.status == "FAILED" for f in dataset.files])



from datetime import datetime
from typing import List

from indico import IndicoClient
from indico.queries.document_report import GetDocumentReport
from indico.types.document_report import DocumentReport
from indico.filters import DocumentReportFilter


def test_fetch_docs(indico):
    client = IndicoClient()

    document_report: List[DocumentReport] = client.call(GetDocumentReport())
    assert document_report is not None
    assert len(document_report) > 0
    assert isinstance(document_report[0].input_files[0].num_pages, int)
    assert isinstance(document_report[0].input_files[0].file_size, int)


def test_fetch_docs_limit(indico):
    client = IndicoClient()
    filter_opts = DocumentReportFilter(
        created_at_start_date=datetime(2021, 7, 1), created_at_end_date=datetime.now()
    )
    document_report: List[DocumentReport] = client.call(
        GetDocumentReport(filters=filter_opts, limit=5)
    )
    assert document_report is not None
    assert len(document_report) > 0
    assert len(document_report) <= 5


def test_pagination(indico):
    client = IndicoClient()
    document_report: List[DocumentReport] = []
    for page in client.paginate(GetDocumentReport(limit=10)):
        document_report.extend(page)
    assert document_report is not None
    assert len(document_report) > 0



import json
from pathlib import Path
from typing import List

import pytest

from build.lib.indico.queries.questionnaire import AddLabels
from indico import IndicoConfig
from indico.client import IndicoClient
from indico.queries import JobStatus, RetrieveStorageObject, CreateDataset, GetDataset, ListWorkflows, \
    UpdateWorkflowSettings, WorkflowSubmission, SubmissionResult, UpdateSubmission, GetSubmission, WaitForSubmissions, \
    SubmitReview
from indico.queries.questionnaire import CreateQuestionaire, GetQuestionnaireExamples, GetQuestionnaire
from indico.types.workflow_metrics import WorkflowMetrics, WorkflowMetricsOptions
from indico.queries.workflow_metrics import GetWorkflowMetrics
from datetime import datetime
from ..data.datasets import *  # noqa
from ..data.datasets import PUBLIC_URL
import time

@pytest.fixture
def workflow(indico, org_annotate_dataset, org_annotate_model_group: ModelGroup):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[org_annotate_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)
    wf = client.call(
        UpdateWorkflowSettings(wf, enable_review=True, enable_auto_review=True)
    )
    assert wf.review_enabled and wf.auto_review_enabled

    _file = str(Path(__file__).parents[1]) + "/data/org-sample.pdf"

    sub_ids = client.call(WorkflowSubmission(workflow_id=wf.id, files=[_file]))
    subs = client.call(WaitForSubmissions(sub_ids, timeout=900))
    sub = subs[0]
    assert sub.status == "PENDING_AUTO_REVIEW"
    raw_result = client.call(RetrieveStorageObject(sub.result_file))
    changes = raw_result["results"]["document"]["results"]
    for model, preds in changes.items():
        if isinstance(preds, dict):
            preds["accepted"] = True
        elif isinstance(preds, list):
            for pred in preds:
                pred["accepted"] = True
    job = client.call(
        SubmitReview(sub.id, changes=changes, force_complete=True)
    )
    job = client.call(JobStatus(job.id, timeout=900))
    submission = client.call(GetSubmission(sub.id))
    assert submission.status == "COMPLETE"

    return wf

def test_fetch_metrics(indico, org_annotate_dataset, workflow):
    client = IndicoClient()
    #time.sleep(300)
    workflow_metric: List[WorkflowMetrics] = \
        client.call(GetWorkflowMetrics(options=[WorkflowMetricsOptions.SUBMISSIONS], start_date=datetime.now(),
                                       end_date=datetime.now(), workflow_ids=[workflow.id]))
    assert workflow_metric is not None
    assert workflow_metric[0].submissions is not None
    assert workflow_metric[0].submissions.aggregate.submitted is 1



from typing import List
from indico.types.base import BaseType


class PerClassSeqMetrics(BaseType):
    """
    Metrics per class

    Attributes:
        span_type (str): Type of span the metric is evaluated on: token, superset, overlap, exact
        f1_score (float): F1 score per class
        precision (float): precision per class
        recall (float): recall per class
        true_positives (int): true positives per class
        false_negatives (int): false negatives per class
        false_positives (int): false_positives
    """

    span_type: str
    f1_score: float
    precision: float
    recall: float
    true_positives: int
    false_negatives: int
    false_positives: int


class AnnotationClassMetrics(BaseType):
    """
    Sequence model metrics per class

    Attributes:
        name (str): class name
        metrics List[PerClassMetrics]: Object containing metrics per class
    """

    name: str
    metrics: List[PerClassSeqMetrics]


class ModelLevelMetrics(BaseType):
    """
    Sequence model metrics at the model level

    Attributes:
        span_type (str): Type of span the metric is evaluated on: token, superset, overlap, exact
        macro_f1 (float): f1-score calculated by taking simple average of class-specific scores
        micro_f1 (float): f1-score calculated by weighting instances across classes
        weighted_f1 (float): f1-score calculated by taking class-specific f1-scores and weighting by available
        positive examples per class
    """

    span_type: str
    macro_f1: float
    micro_f1: float
    weighted_f1: float


class SequenceMetrics(BaseType):
    """
    Model performance metrics calculated for sequence or "annotation" models.

    Attributes:

        class_metrics List[AnnotationClassMetrics]: List of AnnotationClassMetrics objects per class
        model_level_metrics List[ModelLevelMetrics]: List of ModelLevelMetrics objects per span type
        retrain_for_metrics bool: Older annotation models require retraining metrics calculation
    """

    class_metrics: List[AnnotationClassMetrics]
    model_level_metrics: List[ModelLevelMetrics]
    retrain_for_metrics: bool



import re


_cc_to_snake_re = re.compile(r"(?<!^)(?=[A-Z])")
_snake_to_cc_re = re.compile(r"(.*?)_([a-zA-Z])")


def cc_to_snake(string: str):
    return re.sub(_cc_to_snake_re, "_", string).lower()


def _camel(match):
    return match.group(1) + match.group(2).upper()


def snake_to_cc(string: str):
    return re.sub(_snake_to_cc_re, _camel, string, 0)



import pytest
from indico.client import IndicoClient
from indico.filters import or_, UserMetricsFilter
from indico.queries import JobStatus, RetrieveStorageObject
from indico.types.user_metrics import UserSummary
from indico.queries.usermetrics import GetUserSummary, GetUserSnapshots, GetUserChangelog, GenerateChangelogReport
from datetime import datetime


def test_fetch_summary(indico):
    client = IndicoClient()
    user_summary: UserSummary = client.call(GetUserSummary())
    assert user_summary is not None
    assert len(user_summary.app_roles) > 0


def test_fetch_snapshots(indico):
    client = IndicoClient()
    snapshots = []
    for snapshot in client.paginate(GetUserSnapshots(date=datetime.now())):
        snapshots.extend(snapshot)
    assert snapshots is not None
    assert snapshots[0] is not None
    assert len(snapshots) > 0
    first = snapshots.pop()
    assert first is not None
    assert first.roles is not None


def test_fetch_snapshots_with_limit(indico):
    client = IndicoClient()
    snapshots = client.call(GetUserSnapshots(date=datetime.now(), limit=10))
    assert snapshots is not None
    assert snapshots[0] is not None
    assert len(snapshots) == 10
    first = snapshots.pop()
    assert first is not None
    assert first.roles is not None


def test_fetch_filtered_snapshots(indico):
    client = IndicoClient()
    snapshots = []
    filter_opts = or_(UserMetricsFilter(user_id=1), UserMetricsFilter(user_id=2))
    for snapshot in client.paginate(GetUserSnapshots(date=datetime.now(), filters=filter_opts)):
        snapshots.extend(snapshot)
    assert len(snapshots) == 2


def test_changelog(indico):
    client = IndicoClient()
    changelogs = []
    for log in client.paginate((GetUserChangelog(start_date=datetime.now(), end_date=datetime.now(), limit=100))):
        changelogs.extend(log)
    assert len(changelogs) > 0


def test_csv_changelog(indico):
    client = IndicoClient()
    changelogs = client.call((GenerateChangelogReport(start_date=datetime.now(), end_date=datetime.now())))
    assert changelogs is not None
    job = changelogs.job_id
    assert job is not None
    job = client.call(JobStatus(id=job, wait=True))
    assert job.status == "SUCCESS"
    assert job.ready is True
    result = client.call(RetrieveStorageObject(job.result))
    assert result is not None



import io
from typing import List, Union, Dict

from indico.client.request import GraphQLRequest, RequestChain, Debouncer
from indico.errors import IndicoError, IndicoInputError
from indico.queries.storage import UploadDocument
from indico.types import Job, Submission, Workflow, SUBMISSION_RESULT_VERSIONS
from indico.types.utils import cc_to_snake, snake_to_cc


class ListWorkflows(GraphQLRequest):
    """
    List all workflows visible to authenticated user

    Options:
        dataset_ids (List[int]): List of dataset ids to filter by
        workflow_ids (List[int]): List of workflow ids to filter by

    Returns:
        List[Workflow]: All the found Workflow objects
    """

    query = """
        query ListWorkflows($datasetIds: [Int], $workflowIds: [Int]){
            workflows(datasetIds: $datasetIds, workflowIds: $workflowIds){
                workflows {
                    id
                    name
                    status
                    reviewEnabled
                    autoReviewEnabled
                }
            }
        }
    """

    def __init__(
            self, *, dataset_ids: List[int] = None, workflow_ids: List[int] = None
    ):
        super().__init__(
            self.query,
            variables={"datasetIds": dataset_ids, "workflowIds": workflow_ids},
        )

    def process_response(self, response) -> List[Workflow]:
        return [
            Workflow(**w)
            for w in super().process_response(response)["workflows"]["workflows"]
        ]


class GetWorkflow(ListWorkflows):
    """
    Query for Workflow by id

    Args:
        workflow_id (int): Workflow id to query for

    Returns:
        Workflow: Found Workflow object
    """

    def __init__(self, workflow_id: int):
        super().__init__(workflow_ids=[workflow_id])

    def process_response(self, response) -> Workflow:
        return super().process_response(response)[0]


class _ToggleReview(GraphQLRequest):
    toggle = "enableReview"
    query_name = "toggleWorkflowReview"
    query = """
        mutation ToggleReview($workflowId: Int!, $reviewState: Boolean!){
            <QUERY NAME>(workflowId: $workflowId, <TOGGLE>: $reviewState){
                id
                name
                reviewEnabled
                autoReviewEnabled
            }
        }
    """

    def __init__(self, workflow_id: int, enable_review: bool):
        query = self.query.replace("<QUERY NAME>", self.query_name)
        query = query.replace("<TOGGLE>", self.toggle)
        super().__init__(
            query,
            variables={"workflowId": workflow_id, "reviewState": enable_review},
        )

    def process_response(self, response) -> Workflow:
        return Workflow(**super().process_response(response)[self.query_name])


class _ToggleAutoReview(_ToggleReview):
    toggle = "enableAutoReview"
    query_name = "toggleWorkflowAutoReview"


class UpdateWorkflowSettings(RequestChain):
    """
    Mutation to toggle review and auto-review on a workflow

    Args:
        workflow (int|Workflow): Workflow or workflow id to update

    Options:
        enable_review (bool): Enable review
        enable_auto_review (bool): Enable auto_review

    Returns:
        Workflow: Updated Workflow object
    """

    def __init__(
            self,
            workflow: Union[Workflow, int],
            enable_review: bool = None,
            enable_auto_review: bool = None,
    ):
        self.workflow_id = workflow.id if isinstance(workflow, Workflow) else workflow
        if enable_review is None and enable_auto_review is None:
            raise IndicoInputError("Must provide at least one review option")

        self.enable_review = enable_review
        self.enable_auto_review = enable_auto_review

    def requests(self):
        if self.enable_review is not None:
            yield _ToggleReview(self.workflow_id, self.enable_review)
        if self.enable_auto_review is not None:
            yield _ToggleAutoReview(self.workflow_id, self.enable_auto_review)


class _WorkflowSubmission(GraphQLRequest):
    query = """
        mutation workflowSubmissionMutation({signature}) {{
            {mutation_name}({args}) {{
                jobIds
                submissionIds
            }}
        }}
    """

    detailed_query = """
        mutation workflowSubmissionMutation({signature}) {{
            {mutation_name}({args}) {{
                submissionIds
                submissions {{
                    id
                    datasetId
                    workflowId
                    status
                    <SUBQUERY>
                    inputFile
                    inputFilename
                    resultFile
                    retrieved
                    errors
                }}
            }}
        }}
    """
    files_subquery = """
        inputFiles {{
            filepath
            filename
        }}
    """.strip()

    mutation_name = "workflowSubmission"
    mutation_args = {
        "workflowId": "Int!",
        "files": "[FileInput]!",
        "recordSubmission": "Boolean",
        "bundle": "Boolean",
        "resultVersion": "SubmissionResultVersion",
    }

    def __init__(
            self,
            detailed_response: bool,
            **kwargs,
    ):
        self.workflow_id = kwargs["workflow_id"]
        self.record_submission = kwargs["record_submission"]

        # construct mutation signature and args based on provided kwargs to ensure
        # backwards-compatible graphql calls
        #
        # inputFiles, bundle, and resultVersion only avaliable on IPA 4.9.0+
        subq = (
            self.files_subquery
            if kwargs.get("bundle") or kwargs.get("result_version")
            else ""
        )
        q = (
            self.detailed_query.replace("<SUBQUERY>", subq)
            if detailed_response
            else self.query
        )

        args = [
            _arg for _arg in self.mutation_args.keys() if kwargs.get(cc_to_snake(_arg))
        ]
        signature = ",".join(f"${_arg}: {self.mutation_args[_arg]}" for _arg in args)
        args = ",".join(f"{_arg}: ${_arg}" for _arg in args)

        super().__init__(
            query=q.format(
                mutation_name=self.mutation_name, signature=signature, args=args
            ),
            variables={snake_to_cc(var): val for var, val in kwargs.items()},
        )

    def process_response(self, response):
        response = super().process_response(response)[self.mutation_name]
        if "submissions" in response:
            return [Submission(**s) for s in response["submissions"]]
        elif self.record_submission:
            return response["submissionIds"]
        elif not response["jobIds"]:
            raise IndicoError(f"Failed to submit to workflow {self.workflow_id}")
        return [Job(id=job_id) for job_id in response["jobIds"]]


class _WorkflowUrlSubmission(_WorkflowSubmission):
    mutation_name = "workflowUrlSubmission"
    mutation_args = {**_WorkflowSubmission.mutation_args, "urls": "[String]!"}
    del mutation_args["files"]


class WorkflowSubmission(RequestChain):
    f"""
    Submit files to a workflow for processing.
    One of `files` or `urls` is required.

    Args:
        workflow_id (int): Id of workflow to submit files to
        files (List[str], optional): List of local file paths to submit
        urls (List[str], optional): List of urls to submit
        submission (bool, optional): Process these files as normal submissions.
            Defaults to True.
            If False, files will be processed as AsyncJobs, ignoring any workflow
            post-processing steps like Review and with no record in the system
        bundle (bool, optional): Batch all files under a single submission id
        result_version (str, optional):
            The format of the submission result file. One of:
                {SUBMISSION_RESULT_VERSIONS}
            If bundle is enabled, this must be version TWO or later.
        streams (Dict[str, io.BufferedIOBase]): List of filename keys mapped to streams
            for upload. Similar to files but mutually exclusive with files. 
            Can take for example: io.BufferedReader, io.BinaryIO, or io.BytesIO.

    Returns:
        List[int]: If `submission`, these will be submission ids.
            Otherwise, they will be AsyncJob ids.

    """

    detailed_response = False

    def __init__(
            self,
            workflow_id: int,
            files: List[str] = None,
            urls: List[str] = None,
            submission: bool = True,
            bundle: bool = False,
            result_version: str = None,
            streams: Dict[str, io.BufferedIOBase] = None
    ):
        self.workflow_id = workflow_id
        self.files = files
        self.urls = urls
        self.submission = submission
        self.bundle = bundle
        self.result_version = result_version
        self.has_streams = False
        if streams is not None:
            self.streams = streams.copy()
            self.has_streams = True
        else:
            self.streams = None

        if not self.files and not self.urls and not self.has_streams:
            raise IndicoInputError("One of 'files', 'streams', or 'urls' must be specified")
        elif self.files and self.has_streams:
            raise IndicoInputError("Only one of 'files' or 'streams' or 'urls' may be specified.")
        elif (self.files or self.has_streams) and self.urls:
            raise IndicoInputError("Only one of 'files' or 'streams' or 'urls' may be specified")

    def requests(self):
        if self.files:
            yield UploadDocument(files=self.files)
            yield _WorkflowSubmission(
                self.detailed_response,
                workflow_id=self.workflow_id,
                record_submission=self.submission,
                files=self.previous,
                bundle=self.bundle,
                result_version=self.result_version,
            )
        elif self.urls:
            yield _WorkflowUrlSubmission(
                self.detailed_response,
                workflow_id=self.workflow_id,
                record_submission=self.submission,
                urls=self.urls,
                bundle=self.bundle,
                result_version=self.result_version,
            )
        elif self.has_streams:
            yield UploadDocument(streams=self.streams)
            yield _WorkflowSubmission(
                self.detailed_response,
                workflow_id=self.workflow_id,
                record_submission=self.submission,
                files=self.previous,
                bundle=self.bundle,
                result_version=self.result_version,
            )


class WorkflowSubmissionDetailed(WorkflowSubmission):
    f"""
    Submit files to a workflow for processing.
    One of `files` or `urls` is required.
    Submission recording is mandatory.

    Args:
        workflow_id (int): Id of workflow to submit files to
        files (List[str], optional): List of local file paths to submit
        urls (List[str], optional): List of urls to submit
        bundle (bool, optional): Batch all files under a single submission id
        result_version (str, optional):
            The format of the submission result file. One of:
                {SUBMISSION_RESULT_VERSIONS}
            If bundle is enabled, this must be version TWO or later.

    Returns:
        List[Submission]: Submission objects created

    """

    detailed_response = True

    def __init__(
            self,
            workflow_id: int,
            files: List[str] = None,
            urls: List[str] = None,
            bundle: bool = False,
            result_version: str = None,
    ):
        super().__init__(
            workflow_id,
            files=files,
            urls=urls,
            submission=True,
            bundle=bundle,
            result_version=result_version,
        )


class _AddDataToWorkflow(GraphQLRequest):
    query = """
        mutation addDataToWorkflow($workflowId: Int!) {
            addDataToWorkflow(workflowId: $workflowId){
                workflow{
                    id
                    name
                    status
                }
            }
        }
    """

    def __init__(self, workflow_id: int):
        super().__init__(
            self.query,
            variables={"workflowId": workflow_id},
        )

    def process_response(self, response) -> Workflow:
        return Workflow(
            **super().process_response(response)["addDataToWorkflow"]["workflow"]
        )


class AddDataToWorkflow(RequestChain):
    """
    Mutation to update data in a workflow, pressumably
    after new data is added to the dataset.

    Args:
        workflow_id (int): Workflow id to update

    Options:
        wait (bool, default=False): Block while polling for status of update

    Returns:
        Workflow: Updated Workflow object
    """

    def __init__(self, workflow_id: int, wait: bool = False):
        self.workflow_id = workflow_id
        self.wait = wait

    def requests(self):
        yield _AddDataToWorkflow(self.workflow_id)

        if self.wait:
            debouncer = Debouncer()

            while self.previous.status != "COMPLETE":
                yield GetWorkflow(workflow_id=self.workflow_id)
                debouncer.backoff()



import os
import pytest
import time
from pathlib import Path
from indico.client import IndicoClient
from indico.queries import CreateDataset, CreateModelGroup
from indico.types import ModelGroup, Dataset

PUBLIC_URL = "https://github.com/IndicoDataSolutions/indico-client-python/raw/master/tests/integration/data/"


@pytest.fixture(scope="module")
def airlines_dataset(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[0]) + "/AirlineComplaints.csv"

    response = client.call(
        CreateDataset(
            name=f"AirlineComplaints-test-{int(time.time())}", files=[dataset_filepath]
        )
    )
    assert response.status == "COMPLETE"
    return response


@pytest.fixture(scope="module")
def too_small_dataset(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[0]) + "/TooSmall.csv"

    response = client.call(
        CreateDataset(
            name=f"TooSmall-test-{int(time.time())}", files=[dataset_filepath]
        )
    )
    assert response.status == "COMPLETE"
    return response


@pytest.fixture(scope="module")
def airlines_model_group(indico, airlines_dataset: Dataset) -> ModelGroup:
    client = IndicoClient()
    name = f"TestCreateModelGroup-{int(time.time())}"
    mg: ModelGroup = client.call(
        CreateModelGroup(
            name=name,
            dataset_id=airlines_dataset.id,
            source_column_id=airlines_dataset.datacolumn_by_name("Text").id,
            labelset_id=airlines_dataset.labelset_by_name("Target_1").id,
            wait=True,
        )
    )
    return mg


@pytest.fixture(scope="module")
def cats_dogs_image_dataset(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[0]) + "/dog_vs_cats_small.csv"

    response = client.call(
        CreateDataset(
            name=f"DogsAndCats-test-{int(time.time())}",
            files=dataset_filepath,
            from_local_images=True,
        )
    )
    assert response.status == "COMPLETE"
    return response


@pytest.fixture(scope="module")
def cats_dogs_modelgroup(indico, cats_dogs_image_dataset: Dataset) -> ModelGroup:
    client = IndicoClient()
    name = f"TestCreateObjectDetectionMg-{int(time.time())}"

    model_training_options = {
        "max_iter": 20,
        "lr": 0.1,
        "batch_size": 1,
        "filter_empty": False,
        "test_size": 0.2,
        "use_small_model": True,
    }

    mg: ModelGroup = client.call(
        CreateModelGroup(
            name=name,
            dataset_id=cats_dogs_image_dataset.id,
            source_column_id=cats_dogs_image_dataset.datacolumn_by_name("urls").id,
            labelset_id=cats_dogs_image_dataset.labelset_by_name("label").id,
            model_training_options=model_training_options,
            wait=True,
        )
    )
    return mg


@pytest.fixture(scope="module")
def org_annotate_dataset(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[0]) + "/org-annotate-labeled.csv"

    response = client.call(
        CreateDataset(
            name=f"OrgAnnotate-test-{int(time.time())}", files=[dataset_filepath]
        )
    )
    assert response.status == "COMPLETE"
    return response


@pytest.fixture(scope="module")
def org_annotate_model_group(indico, org_annotate_dataset: Dataset) -> ModelGroup:
    client = IndicoClient()
    name = f"TestFinetuneModelGroup-{int(time.time())}"
    mg: ModelGroup = client.call(
        CreateModelGroup(
            name=name,
            dataset_id=org_annotate_dataset.id,
            source_column_id=org_annotate_dataset.datacolumn_by_name(
                "News Headlines w/Company Names"
            ).id,
            labelset_id=org_annotate_dataset.labelset_by_name("question_825").id,
            wait=True,
        )
    )
    return mg



import logging
from indico._version import get_versions

logging.basicConfig()
Version = version = __version__ = VERSION = get_versions()["version"]

from indico.client import *
from indico.client.request import *

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions



from typing import List

from indico.types import BaseType


class InputFile(BaseType):
    filename: str
    submission_id: int
    num_pages: int
    file_size: int


class DocumentReport(BaseType):
    """
    A Document report about the associated InputFiles.


    """
    dataset_id: int
    workflow_id: int
    status: str
    created_at: str
    updated_at: str
    updated_by: str
    completed_at: str
    errors: str
    retrieved: bool
    input_files: List[InputFile]




from indico.queries.workflow import GetWorkflow
import pytest
from pathlib import Path
import time

from indico.client import IndicoClient
from indico.errors import IndicoError, IndicoInputError
from indico.filters import SubmissionFilter
from indico.queries import (
    GetSubmission,
    JobStatus,
    ListSubmissions,
    ListWorkflows,
    RetrieveStorageObject,
    SubmissionResult,
    SubmitReview,
    UpdateSubmission,
    UpdateWorkflowSettings,
    WaitForSubmissions,
    WorkflowSubmission,
    WorkflowSubmissionDetailed,
    AddDataToWorkflow,
    AddFiles,
    ProcessCSV,
    CreateDataset,
)
from indico.queries.questionnaire import CreateQuestionaire, GetQuestionnaire
from indico.types import ModelGroup
from indico.types.submission import Submission

from ..data.datasets import *  # noqa
from ..data.datasets import PUBLIC_URL


def test_list_workflows(indico, airlines_dataset, airlines_model_group: ModelGroup):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    assert len(wfs) > 0


def test_workflow_job(indico, airlines_dataset, airlines_model_group: ModelGroup):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)

    dataset_filepath = str(Path(__file__).parents[1]) + "/data/mock.pdf"

    jobs = client.call(
        WorkflowSubmission(
            workflow_id=wf.id, files=[dataset_filepath], submission=False
        )
    )
    job = jobs[0]

    assert job.id is not None
    job = client.call(JobStatus(id=job.id, wait=True))
    assert job.status == "SUCCESS"
    assert job.ready is True
    assert isinstance(job.result["url"], str)

    result = client.call(RetrieveStorageObject(job.result))

    assert isinstance(result, dict)


@pytest.mark.parametrize(
    "_input",
    [
        {"urls": [PUBLIC_URL + "mock.pdf"]},
        {"files": [str(Path(__file__).parents[1]) + "/data/mock.pdf"]},
    ],
)
def test_workflow_submission(
    indico, airlines_dataset, airlines_model_group: ModelGroup, _input
):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)

    submission_ids = client.call(WorkflowSubmission(workflow_id=wf.id, **_input))
    submission_id = submission_ids[0]
    assert submission_id is not None

    with pytest.raises(IndicoInputError):
        client.call(SubmissionResult(submission_id, "FAILED"))

    with pytest.raises(IndicoInputError):
        client.call(SubmissionResult(submission_id, "INVALID_STATUS"))

    result_url = client.call(SubmissionResult(submission_id, "COMPLETE", wait=True))
    result = client.call(RetrieveStorageObject(result_url.result))
    assert isinstance(result, dict)
    assert result["submission_id"] == submission_id
    assert result["file_version"] == 1
    sub = client.call(GetSubmission(submission_id))
    assert isinstance(sub, Submission)
    assert sub.retrieved is False
    assert sub.deleted is False
    client.call(UpdateSubmission(submission_id, retrieved=True))
    sub = client.call(GetSubmission(submission_id))
    assert sub.retrieved is True

def test_workflow_submission_with_streams(
    indico, airlines_dataset, airlines_model_group: ModelGroup
):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)

    path = Path(str(Path(__file__).parents[1]) + "/data/mock.pdf")
    fd = open(path.absolute(), "rb")
    files = {
        "mock.pdf": fd
    }
    submission_ids = client.call(WorkflowSubmission(workflow_id=wf.id, streams=files))
    submission_id = submission_ids[0]
    assert submission_id is not None

    with pytest.raises(IndicoInputError):
        client.call(SubmissionResult(submission_id, "FAILED"))

    with pytest.raises(IndicoInputError):
        client.call(SubmissionResult(submission_id, "INVALID_STATUS"))

    result_url = client.call(SubmissionResult(submission_id, "COMPLETE", wait=True))
    result = client.call(RetrieveStorageObject(result_url.result))
    assert isinstance(result, dict)
    assert result["submission_id"] == submission_id
    assert result["file_version"] == 1
    client.call(UpdateSubmission(submission_id, retrieved=True))
    sub = client.call(GetSubmission(submission_id))
    assert isinstance(sub, Submission)
    assert sub.retrieved is True



@pytest.mark.parametrize(
    "_input",
    [
        {"urls": [PUBLIC_URL + "mock.pdf"] * 3},
        {"files": [str(Path(__file__).parents[1]) + "/data/mock.pdf"] * 3},
    ],
)
def test_workflow_submission_versioned(
    indico, airlines_dataset, airlines_model_group: ModelGroup, _input
):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)

    submission_ids = client.call(
        WorkflowSubmission(workflow_id=wf.id, result_version="LATEST", **_input)
    )

    assert len(submission_ids) == len(next(iter(_input.values())))
    submission_id = submission_ids[0]
    assert submission_id is not None

    submissions = client.call(WaitForSubmissions(submission_id))
    result = client.call(RetrieveStorageObject(submissions[0].result_file))

    assert isinstance(result, dict)
    assert result["file_version"] == 2
    assert len(result["submission_results"]) == 1
    assert result["submission_results"][0]["input_filename"] == "mock.pdf"


@pytest.mark.parametrize(
    "_input",
    [
        {"urls": [PUBLIC_URL + "mock.pdf"] * 3},
        {"files": [str(Path(__file__).parents[1]) + "/data/mock.pdf"] * 3},
    ],
)
def test_workflow_submission_bundled(
    indico, airlines_dataset, airlines_model_group: ModelGroup, _input
):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)

    submission_ids = client.call(
        WorkflowSubmission(
            workflow_id=wf.id, bundle=True, result_version="LATEST", **_input
        )
    )

    assert len(submission_ids) == 1
    submission_id = submission_ids[0]
    assert submission_id

    submissions = client.call(WaitForSubmissions(submission_id))
    result = client.call(RetrieveStorageObject(submissions[0].result_file))

    assert isinstance(result, dict)
    assert result["file_version"] == 2
    assert len(result["submission_results"]) == len(next(iter(_input.values())))
    assert result["submission_results"][0]["input_filename"] == "mock.pdf"


@pytest.mark.parametrize(
    "_input,_output",
    [
        ({"urls": [PUBLIC_URL + "mock.pdf"]}, PUBLIC_URL + "mock.pdf"),
        ({"files": [str(Path(__file__).parents[1]) + "/data/mock.pdf"]}, "mock.pdf"),
    ],
)
def test_workflow_submission_detailed(
    indico, airlines_dataset, airlines_model_group: ModelGroup, _input, _output
):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)

    submissions = client.call(WorkflowSubmissionDetailed(workflow_id=wf.id, **_input))
    assert isinstance(submissions[0], Submission)
    assert submissions[0].input_filename == _output


def test_list_workflow_submission_retrieved(
    indico, airlines_dataset, airlines_model_group: ModelGroup
):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)

    dataset_filepath = str(Path(__file__).parents[1]) + "/data/mock.pdf"

    submission_ids = client.call(
        WorkflowSubmission(workflow_id=wf.id, files=[dataset_filepath])
    )
    submission_id = submission_ids[0]
    assert submission_id is not None
    client.call(SubmissionResult(submission_id, "COMPLETE", wait=True))
    client.call(UpdateSubmission(submission_id, retrieved=True))

    submissions = client.call(
        ListSubmissions(filters=SubmissionFilter(retrieved=True, status="COMPLETE"))
    )
    assert all([s.retrieved for s in submissions])
    assert submission_id in [s.id for s in submissions]

    submissions = client.call(
        ListSubmissions(filters=SubmissionFilter(retrieved=False))
    )
    assert all([not s.retrieved for s in submissions])
    assert submission_id not in [s.id for s in submissions]

def test_list_workflow_submission_paginate(
    indico, airlines_dataset, airlines_model_group: ModelGroup
):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)

    dataset_filepath = str(Path(__file__).parents[1]) + "/data/mock.pdf"

    submission_ids = client.call(
        WorkflowSubmission(workflow_id=wf.id, files=[dataset_filepath]*5)
    )
    for sub in client.paginate(ListSubmissions(workflow_ids=[wf.id], limit=3)):

        if not submission_ids:
            break
        assert sub.id == submission_ids.pop()  # list is desc by default

def test_workflow_submission_missing_workflow(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[1]) + "/data/mock.pdf"

    with pytest.raises(IndicoError):
        client.call(WorkflowSubmission(workflow_id=0, files=[dataset_filepath]))


def test_workflow_submission_mixed_args(indico, airlines_dataset):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[airlines_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)

    _file = str(Path(__file__).parents[1]) + "/data/mock.pdf"
    url = PUBLIC_URL + "mock.pdf"

    with pytest.raises(IndicoInputError):
        client.call(WorkflowSubmission(workflow_id=wf.id, files=[_file], urls=[url]))


def test_auto_review_enabled_in_get_workflow_response(
    indico, org_annotate_dataset, org_annotate_model_group
):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[org_annotate_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)

    assert not (wf.auto_review_enabled or wf.review_enabled)

    wf = client.call(
        UpdateWorkflowSettings(wf, enable_review=True, enable_auto_review=True)
    )
    assert wf.review_enabled and wf.auto_review_enabled

    wf = client.call(GetWorkflow(workflow_id=wf.id))
    assert wf.review_enabled and wf.auto_review_enabled


@pytest.mark.parametrize("force_complete", [None, True])
def test_workflow_submission_auto_review(
    indico, force_complete, org_annotate_dataset, org_annotate_model_group
):
    client = IndicoClient()
    wfs = client.call(ListWorkflows(dataset_ids=[org_annotate_dataset.id]))
    wf = max(wfs, key=lambda w: w.id)
    wf = client.call(
        UpdateWorkflowSettings(wf, enable_review=True, enable_auto_review=True)
    )
    assert wf.review_enabled and wf.auto_review_enabled

    _file = str(Path(__file__).parents[1]) + "/data/org-sample.pdf"

    sub_ids = client.call(WorkflowSubmission(workflow_id=wf.id, files=[_file]))
    subs = client.call(WaitForSubmissions(sub_ids, timeout=120))
    sub = subs[0]
    assert sub.status == "PENDING_AUTO_REVIEW"
    raw_result = client.call(RetrieveStorageObject(sub.result_file))
    changes = raw_result["results"]["document"]["results"]
    for model, preds in changes.items():
        if isinstance(preds, dict):
            preds["accepted"] = True
        elif isinstance(preds, list):
            for pred in preds:
                pred["accepted"] = True
    job = client.call(
        SubmitReview(sub.id, changes=changes, force_complete=force_complete)
    )
    job = client.call(JobStatus(job.id))
    submission = client.call(GetSubmission(sub.id))
    assert submission.status == "COMPLETE" if force_complete else "PENDING_REVIEW"


def _new_dataset_for_updating(client):
    # new dataset
    airline_csv = str(Path(__file__).parents[1]) + "/data/AirlineComplaints.csv"
    dataset = client.call(
        CreateDataset(
            name=f"AddDataToWorkflow-test-{int(time.time())}", files=[airline_csv]
        )
    )

    # new teach task
    questionnaire = client.call(
        CreateQuestionaire(
            name=f"AddDataToWorkflowTeach-test-{int(time.time())}",
            dataset_id=dataset.id,
            targets=["positive", "negative"],
        )
    )

    assert questionnaire.num_total_examples > 0

    # add data to dataset and process
    dataset = client.call(AddFiles(dataset_id=dataset.id, files=[airline_csv]))
    datafile_ids = [f.id for f in dataset.files]
    dataset = client.call(
        ProcessCSV(dataset_id=dataset.id, datafile_ids=datafile_ids, wait=True)
    )

    assert dataset.status == "COMPLETE"

    # get workflow
    wfs = client.call(ListWorkflows(dataset_ids=[dataset.id]))
    wf = max(wfs, key=lambda w: w.id)
    return dataset, wf, questionnaire


def test_add_data_to_workflow_wait(indico):
    client = IndicoClient()
    _, wf, q1 = _new_dataset_for_updating(client)

    wf = client.call(AddDataToWorkflow(wf.id, wait=True))
    assert wf.status == "COMPLETE"

    q2 = client.call(GetQuestionnaire(q1.id))
    # the total num avaliable should double, since we re-add the same data
    assert q1.num_total_examples * 2 == q2.num_total_examples


def test_add_data_to_workflow_nowait(indico):
    client = IndicoClient()
    _, wf, _ = _new_dataset_for_updating(client)

    wf = client.call(AddDataToWorkflow(wf.id))
    assert wf.status == "ADDING_DATA"



from typing import List

from indico.types.base import BaseType


class Export(BaseType):
    """
    An export of a dataset.

    Attributes:
        id (int): Id of the export
        dataset_id (int): Dataset id of the export
        status (str): Export job status
        column_ids (List(int)): Data columns for export
        labelset_ids (List(int)): Labelsets columns for export
        subset_ids (List(int)): Subsets requested for export rows
        num_labels (int): Number of labels on the dataset
        anonymous (bool): Whether to anonymize labelers
        download_url (str): Indico url of the export csv
        created_at (str): Unix timestamp for when the export was created
    """

    id: int
    dataset_id: int
    name: str
    status: str
    column_ids: List[int]
    labelset_ids: List[int]
    subset_ids: List[int]
    num_labels: int
    anonymous: bool
    download_url: str
    created_at: str



import os
from pathlib import Path
from indico.client import IndicoClient
from indico.queries import (
    RetrieveStorageObject,
    JobStatus,
    DocumentExtraction,
    UploadBatched,
    UploadDocument,
)


def test_document_extraction(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[1]) + "/data/mock.pdf"

    jobs = client.call(DocumentExtraction(files=[dataset_filepath]))

    assert len(jobs) == 1
    job = jobs[0]
    assert job.id is not None
    job = client.call(JobStatus(id=job.id, wait=True))
    assert job.status == "SUCCESS"
    assert job.ready is True
    assert type(job.result["url"]) == str

    extract = client.call(RetrieveStorageObject(job.result))

    assert type(extract) == dict
    assert "pages" in extract


def test_document_extraction_with_config(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[1]) + "/data/mock.pdf"

    jobs = client.call(
        DocumentExtraction(
            files=[dataset_filepath], json_config={"preset_config": "simple"}
        )
    )

    assert len(jobs) == 1
    job = jobs[0]
    assert job.id is not None
    job = client.call(JobStatus(id=job.id, wait=True))
    assert job.status == "SUCCESS"
    assert job.ready is True
    assert type(job.result["url"]) == str

    extract = client.call(RetrieveStorageObject(job.result))

    assert type(extract) == dict
    assert "pages" in extract


def test_document_extraction_with_string_config(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[1]) + "/data/mock.pdf"

    jobs = client.call(
        DocumentExtraction(
            files=[dataset_filepath], json_config='{"preset_config": "simple"}'
        )
    )

    assert len(jobs) == 1
    job = jobs[0]
    assert job.id is not None
    job = client.call(JobStatus(id=job.id, wait=True))
    assert job.status == "SUCCESS"
    assert job.ready is True
    assert type(job.result["url"]) == str

    extract = client.call(RetrieveStorageObject(job.result))
    assert type(extract) == dict
    assert "pages" in extract


def test_upload_documents_batched(indico):
    file_names = ["mock.pdf", "mock_2.pdf", "mock_3.pdf"]
    client = IndicoClient()
    parent_path = str(Path(__file__).parent.parent / "data")
    dataset_filepaths = [
        os.path.join(parent_path, file_name) for file_name in file_names
    ]

    files = client.call(
        UploadBatched(files=dataset_filepaths, batch_size=1, request_cls=UploadDocument)
    )
    assert len(files) == 3
    for file_name, file in zip(file_names, files):
        assert file["filename"] == file_name


def test_document_extraction_batched(indico):
    client = IndicoClient()
    file_names = ["mock.pdf", "mock_2.pdf", "mock_3.pdf"]
    parent_path = str(Path(__file__).parent.parent / "data")
    dataset_filepaths = [
        os.path.join(parent_path, file_name) for file_name in file_names
    ]

    jobs = client.call(
        DocumentExtraction(
            files=dataset_filepaths,
            json_config={"preset_config": "simple"},
            upload_batch_size=1,
        )
    )
    assert len(jobs) == 3
    for job in jobs:
        assert job.id is not None
        job = client.call(JobStatus(id=job.id, wait=True))
        assert job.status == "SUCCESS"
        assert job.ready is True
        assert isinstance(job.result["url"], str)


def test_document_extraction_thumbnails(indico):
    client = IndicoClient()
    dataset_filepath = str(Path(__file__).parents[1]) + "/data/mock.pdf"

    jobs = client.call(DocumentExtraction(files=[dataset_filepath]))

    assert len(jobs) == 1
    job = jobs[0]
    assert job.id is not None
    job = client.call(JobStatus(id=job.id, wait=True))
    assert job.status == "SUCCESS"
    assert job.ready is True
    assert type(job.result["url"]) == str

    extract = client.call(RetrieveStorageObject(job.result))

    assert type(extract) == dict
    assert "pages" in extract
    image = extract["pages"][0]["image"]

    image = client.call(RetrieveStorageObject(image))

    assert image


def test_upload_duplicate_documents(indico):
    client = IndicoClient()
    file_names = ["mock.pdf", "mock.pdf", "mock_2.pdf"]
    parent_path = str(Path(__file__).parent.parent / "data")
    filepaths = [os.path.join(parent_path, file_name) for file_name in file_names]
    uploaded_files = client.call(UploadDocument(files=filepaths))
    assert len(uploaded_files) == 3
    assert [f["filename"] for f in uploaded_files] == file_names



from enum import Enum
from indico.types.base import BaseType
from typing import List


class WorkflowMetricsOptions(Enum):
    """
    Options for pulling back specific metrics.

    Attributes:
        SUBMISSIONS: Generates SubmissionMetrics in the response.
        REIVEW: Generates review Queue metrics in the response.
        STRAIGHT_THROUGH_PROCESSING: Generates StraightThroughProcessing metrics in the response.
        TIME_ON_TASK: Average time spent by reviewers on documents.
    """
    SUBMISSIONS = 1
    REVIEW = 2
    STRAIGHT_THROUGH_PROCESSING = 3
    TIME_ON_TASK = 4


class TimeOnTaskMetric(BaseType):
    """
    Time on task metrics.

    Attributes:
        avg_mins_per_doc(float): Average amount of minutes reviewers spend on documents in this workflow
            across review and exceptions queue.
        avg_mins_per_doc_review(float): Average amount of minutes reviewers spend on documents in this workflow
            in the review queue.
        avg_min_per_doc_exceptions(float): Average amount of minutes reviewers spend on ducments for this workflow in
            the exceptions queue.
    """
    avg_mins_per_doc: float
    avg_mins_per_doc_review: float
    avg_min_per_doc_exceptions: float


class DailyTimeOnTaskMetric(TimeOnTaskMetric):
    """
    Daily time on task metrics.
    
    Attributes:
        avg_mins_per_doc(float): Average amount of minutes reviewers spend on documents in this workflow
            across review and exceptions queue.
        avg_mins_per_doc_review(float): Average amount of minutes reviewers spend on documents in this workflow
            in the review queue.
        avg_min_per_doc_exceptions(float): Average amount of minutes reviewers spend on ducments for this workflow in
            the exceptions queue.
        date(str): Date.
    """
    date: str


class TimeOnTaskMetrics(BaseType):
    """
    Time on task metrics.

    Attributes:
        aggregate(TimeOnTaskMetric): Aggregate level time on task metrics.
        daily(List[DailyTimeOnTaskMetric): Daily level time on task metrics.
    """
    aggregate: TimeOnTaskMetric
    daily: List[DailyTimeOnTaskMetric]


class SubmissionMetric(BaseType):
    """
    Submission metric.

     Attributes:
        submitted(int): Number of items submitted to this workflow on this day.
        completed(int): Number of items completed in this workflow on this day.
        completed_in_review(int): Number of items accepted in the review or the exceptions queue.
        completed_review_queue(int): Number of items accepted in the review queue.
        completed_exception_queue(int): Number of items accepted in the exceptions queue.
        rejected_in_review(int): Number of items rejected in the exceptions queue.
        date(str): Date.
    """
    submitted: int
    completed: int
    completed_in_review: int
    completed_review_queue: int
    completed_exception_queue: int
    rejected_in_review: int


class DailySubmissionMetric(SubmissionMetric):
    """
    Daily submission metrics.

    Attributes:
        submitted(int): Number of items submitted to this workflow on this day.
        completed(int): Number of items completed in this workflow on this day.
        completed_in_review(int): Number of items accepted in the review or the exceptions queue.
        completed_review_queue(int): Number of items accepted in the review queue.
        completed_exception_queue(int): Number of items accepted in the exceptions queue.
        rejected_in_review(int): Number of items rejected in the exceptions queue.
        date(str): Date.
    """
    date: str


class SubmissionMetrics(BaseType):
    """
    Submission metrics.

    Attributes:
        aggregate(SubmissionMetric): Aggregate level submission metrics.
        daily(List[SubmissionMetric]): Daily submission metrics.
    """
    aggregate: SubmissionMetric
    daily: List[DailySubmissionMetric]


class DailyQueueMetric(BaseType):
    """
    Daily cumulative metrics for queues
    Attributes:
        date (str): the date.
        subs_on_queue(int): Number of submissions queued for review
        hours_on_queue(float): Cumulative hours items wait on queue for review.
        avg_age_in_queue(float): Average cumulative age of items waiting in review queues.
    """
    date: str
    subs_on_queue: int
    hours_on_queue: float
    avg_age_in_queue: float


class QueueMetrics(BaseType):
    """
    Daily cumulative metrics for queues.
    Attributes:
        daily_cumulative(List[DailyQueueMetric]): list of cumulative queue metrics per day
    """
    daily_cumulative: List[DailyQueueMetric]


class PredictionMetric(BaseType):
    """
    The total number of model-generated predictions for specified workflow.
    Attributes:
        num_preds(int): the total number of model-generated predictions.
    """
    num_preds: int


class DailyPredictionMetric(PredictionMetric):
    """
    The total number of model-generated predictions for specified workflow, for this date.
    Attributes:
        date(str): the date of the prediction.
        num_preds(int): the total number of model-generated predictions.
    """
    date: str


class PredictionMetrics(BaseType):
    """
    Prediction metrics.

    Attributes:
        aggregate(PredictionMetric): Total number of predictions generated for this workflow.
        daily(List[PredictionMetrics]): Number of predictions generated per-date.
    """
    aggregate: PredictionMetric
    daily: List[PredictionMetric]


class StpMetric(BaseType):
    """
    Base Straight through processing (STP) metric for a model, class, or workflow.

    Attributes:
        review_numerator(int): The number of human accepted model predictions.
        auto_review_numerator(int): The nymber of human accepted auto review labels.
        review_denom(int): Total of user supplied labels and model predictions.
        auto_review_denom(int) Total of user supplied labels and auto review labels.
        review_stp_pct(float): Percent of human accepted model predictions. Present if auto review is disabled.
        auto_review_stp_pct(float): Percent of auto review labels accepted. Present if auto review is enabled.
    """
    review_numerator: int
    auto_review_numerator: int
    review_denom: int
    auto_review_denom: int
    review_stp_pct: float
    auto_review_stp_pct: float


class DailyStpMetric(StpMetric):
    """
    Daily Straight through processing (STP) metric for a model, class, or workflow.

    Attributes:
        review_numerator(int): The number of human accepted model predictions.
        auto_review_numerator(int): The nymber of human accepted auto review labels.
        review_denom(int): Total of user supplied labels and model predictions.
        auto_review_denom(int) Total of user supplied labels and auto review labels.
        review_stp_pct(float): Percent of human accepted model predictions. Present if auto review is disabled.
        auto_review_stp_pct(float): Percent of auto review labels accepted. Present if auto review is enabled.
        date(str): The date these metrics are applicable.
    """
    date: str


class ClassStpMetrics(BaseType):
    """
    STP Metrics per class.

    Attributes:
        class_name(str): Name of the class.
        aggregate(StpMetric): Aggregate level metrics about this class.
        daily(List[DailyStpMetric]): Per-date STP metrics for this class.
    """
    class_name: str
    aggregate: StpMetric
    daily: List[DailyStpMetric]


class ModelStpMetrics(BaseType):
    """
    Model STP metrics.

    Attributes:
        model_group_id(int): Id of the model group.
        name(str): Name of the model.
        aggregate(StpMetric): Aggregate level STP metrics for the model.
        daily(List[DailyStpMetric]): Daily STP metrics for the model.
        class_metrics(List[ClassStpMetrics]): Metrics per model class.
    """
    model_group_id: int
    name: str
    aggregate: StpMetric
    daily: List[DailyStpMetric]
    class_metrics: List[ClassStpMetrics]


class WorkflowStpMetrics(BaseType):
    """
    Daily workflow STP metrics.

    Attributes:
        daily(List[DailyStpMetric]): List of daily metrics.
    """
    daily: List[DailyStpMetric]


class StraightThroughProcessing(BaseType):
    """
    Straight Through Processing (STP) metrics.

    Attributes:
        workflow(WorkflowStpMetrics): Daily aggregate workflow level STP metrics.
        model(List[ModelStpMetrics]): Model STP metrics (including class STP).
    """
    workflow: WorkflowStpMetrics
    model: List[ModelStpMetrics]


class WorkflowMetrics(BaseType):
    """
    Workflow metrics for the requested workflow id and dates.

    Attributes:
        workflow_id(int): The workflow id.
        time_on_task(TimeOnTaskMetrics): Time spent on tasks.
        submissions(SubmissionMetrics): Submission specific metrics.
        queues(QueueMetrics): Review queue metrics.
        predictions(PredictionMetrics): Prediction metrics.
        straight_through_processing(StraightThroughProcessing): Straight through processing (STP) metrics.
        first_submitted_date(str): The earliest date of submission to this workflow.

    """
    workflow_id: int
    time_on_task: TimeOnTaskMetrics
    submissions: SubmissionMetrics
    queues: QueueMetrics
    predictions: PredictionMetrics
    straight_through_processing: StraightThroughProcessing
    first_submitted_date: str



from typing import Dict, Any
from enum import Enum
from indico.errors import IndicoRequestError
import time


class HTTPMethod(Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"


class HTTPRequest:
    def __init__(self, method: HTTPMethod, path: str, **kwargs):
        self.method = method
        self.path = path
        self.kwargs = kwargs

    def process_response(self, response):
        return response


class GraphQLRequest(HTTPRequest):
    def __init__(self, query: str, variables: Dict[str, Any] = None):
        self.query = query
        self.variables = variables
        self.method = HTTPMethod.POST
        self.path = "/graph/api/graphql"

    @property
    def kwargs(self):
        return {"json": {"query": self.query, "variables": self.variables}}

    def process_response(self, response):
        response = super().process_response(response)
        errors = response.pop("errors", [])
        if errors:
            extras = {"locations": [error.pop("locations", None) for error in errors]}
            raise IndicoRequestError(
                error="\n".join(error["message"] for error in errors),
                code=400,
                extras=extras,
            )
        return response["data"]


class PagedRequest(GraphQLRequest):
    """
    To enable pagination, query must include $after as an argument
    and request pageInfo
        query Name(
            ...
            $after: Int
        ){
            items(
                ...
                after: $after
            ){
                items {...}
                pageInfo {
                    endCursor
                    hasNextPage
                }
            }
        }
    """

    def __init__(self, query: str, variables: Dict[str, Any] = None):
        variables["after"] = None
        self.has_next_page = True
        super().__init__(query, variables=variables)

    def process_response(self, response):
        response = super().process_response(response)
        _pg = next(iter(response.values()))["pageInfo"]
        self.has_next_page = _pg["hasNextPage"]
        self.variables["after"] = _pg["endCursor"] if self.has_next_page else None
        return response


class RequestChain:
    previous: Any = None
    result: Any = None

    def requests(self):
        pass


class Debouncer:
    def __init__(self, max_timeout: int = 5):
        self.timeout = 0
        self.max_timeout = max_timeout or 5  # prevent None and 0

    def backoff(self):
        self.increment_timeout()
        time.sleep(self.timeout)

    def increment_timeout(self):
        if self.timeout < self.max_timeout:
            self.timeout += 1



import pytest
from unittest import mock
from indico.client import IndicoClient
from indico.types.dataset import Dataset
from indico.types.export import Export
from indico.errors import IndicoRequestError
from indico.queries.export import CreateExport, DownloadExport
from ..data.datasets import airlines_dataset
import re


def test_create_and_download_export(airlines_dataset: Dataset):
    client = IndicoClient()
    export = client.call(CreateExport(dataset_id=airlines_dataset.id, wait=True))
    assert export.status == "COMPLETE"

    csv = client.call(DownloadExport(export.id))

    assert all(c in csv.columns.to_list() for c in ["ID", "Target_1", "Text"])
    assert any(re.match("row_index_[0-9]+", c) for c in csv.columns.to_list())
    assert csv["Text"][0] == "Your service is so bad."
    assert (
        csv["Target_1"][0] == "You are threatening to never to use this airline again"
    )


def test_download_incomplete(indico):
    client = IndicoClient()
    export = Export()
    export.status = "FAILED"
    with pytest.raises(IndicoRequestError) as e:
        client.call(DownloadExport(export=export))
        assert isinstance(e._excinfo, IndicoRequestError)


def test_create_export_no_wait(airlines_dataset: Dataset):
    client = IndicoClient()
    export = client.call(CreateExport(dataset_id=airlines_dataset.id, wait=False))
    assert export.status == "STARTED"



from indico.types import BaseType, List
from . import SubmissionFile

VALID_SUBMISSION_STATUSES = [
    "COMPLETE",
    "FAILED",
    "PENDING_REVIEW",
    "PROCESSING",
    "PENDING_ADMIN_REVIEW",
    "PENDING_AUTO_REVIEW",
]


SUBMISSION_RESULT_VERSIONS = ["ONE", "TWO", "OLDEST_SUPPORTED", "LATEST"]


class SubmissionRetries(BaseType):
    """
    Information about a retried submission.

    Attributes:
        id (int): The id of the retry attempt.
        previous_errors (str): Errors from the previous attempt.
        previous_status (str): The status from the previous attempt.
        retry_errors (str): Errors encountered on this retry.
        submission_id (int): The ID of the submission being retried.
    """
    id: int
    previous_errors: str
    previous_status: str
    retry_errors: str
    submission_id: int


class Submission(BaseType):
    f"""
    A Submission in the Indico Platform.

    Submissions represent a single input which has been sent for processing by a
    specific workflow. A submission consists of SubmissionFiles, generally a PDF.
    Processing broadly consists of an input processor, a series of processors and
    components associated with particular docbots, and finally an output processor
    to generate the result file.

    Attributes:
        id (int): the Submission id
        dataset_id (int): the Dataset id
        workflow_id (int): the Workflow id
        status (str): status of the submission. One of
            {VALID_SUBMISSION_STATUSES}
        input_files (list[SubmissionFile]): the SubmissionFiles for the Submission
        input_file (str): URL of the first input datafile within the Indico Platform.
        input_filename (str): name of the first original file
        result_file (str): URL of the result datafile within the Indico Platform
        retrieved (bool): Whether the submission has been retrieved by a user
            This flag is set manually by users.
        deleted (bool): Whether the submission result has been deleted from the server
        errors (str): Any errors raised while processing the submission
        retries (List[SubmissionRetries]): If requested, information about previous retries of this submission.
    """

    id: int
    dataset_id: int
    workflow_id: int
    status: str
    input_files: List[SubmissionFile]
    input_file: str
    input_filename: str
    result_file: str
    retrieved: bool
    deleted: bool
    errors: str
    retries: List[SubmissionRetries]



# -*- coding: utf-8 -*-

import os

from pathlib import Path
from indico.errors import IndicoInvalidConfigSetting


class IndicoConfig:
    """
    Configuration for the IndicoClient.

    Args:
        host= (str, optional): Indico Platform hostname (ie indico.myco.com)
        api_token_path= (str, optional): Path to the Indico API token file indico_api_token.txt
        api_token= (str, optional): The actual text of the API Token
        verify_ssl= (bool, optional): Verify the host's SSL certificate? Default=True
        requests_params= (dict, optional): Dictionary of requests.Session parameters to set

    Returns:
        IndicoConfig object

    Raises:
        RuntimeError: If api_token_path does not exist.

    Notes:
        By default, the IndicoConfig constructor will set host and api_token_path from the
        INDICO_HOST and INDICO_API_TOKEN_PATH environment variables. If these variable are
        not set then it will use ``app.indico.io`` as the host and will look for the
        indico_api_token.txt file in the user's home directory.

        If your Indico Platform host is configured to use http instead of https or if the
        host has an invalid SSL certificate then be sure to set verify_ssl=False

        The requests_params dictionary should have keys that are valid attributes of the Session
        object from the requests library.
    """

    host: str
    protocol: str
    serializer: str
    api_token_path: str
    api_token: str = None
    verify_ssl: bool = True
    requests_params: dict = None
    _disable_cookie_domain: bool = False

    def __init__(self, **kwargs):
        self.host: str = os.getenv("INDICO_HOST", "app.indico.io")
        self.protocol: str = os.getenv("INDICO_PROTOCOL", "https")
        self.serializer: str = os.getenv("INDICO_SERIALIZER", "msgpack")
        self.api_token_path: str = os.getenv("INDICO_API_TOKEN_PATH", Path.home())

        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                raise IndicoInvalidConfigSetting(key)

        if not self.api_token:
            self.api_token_path, self.api_token = self._resolve_api_token()

    def _resolve_api_token(self):
        path = self.api_token_path
        if not isinstance(path, Path):
            path = Path(path)
        if not path.exists():
            path = Path.home()
        if not path.is_file():
            path = path / "indico_api_token.txt"

        if not path.exists():
            raise RuntimeError(
                "Expected indico_api_token.txt in home directory, "
                "or provided as indicoio.config.api_token_path"
            )

        with path.open("r") as f:
            return path, f.read().strip()



from typing import List
from indico.types.base import BaseType


class Questionnaire(BaseType):
    id: int
    questions_status: str
    odl: bool
    name: str
    num_total_examples: int
    num_fully_labeled: int


class Example(BaseType):
    row_index: int
    datafile_id: int
    source: str


class QuestionnaireExamples(BaseType):
    examples: List[Example]



import datetime
from typing import Any, Iterable, Mapping

from indico.errors import IndicoInputError


def or_(*args: Iterable[Mapping[str, Any]]):
    return {"OR": list(args)}


def and_(*args: Iterable[Mapping[str, Any]]):
    return {"AND": list(args)}


class Filter(dict):
    """
    Base filter class that allows users to construct filter statements for
    GraphQL queries. Search keys are constrained by the implementing subclasses
    If multiple arguments are supplied, they are treated as arg1 AND arg2 AND ...
    """

    __options__ = None

    def __init__(self, **kwargs):
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        if not kwargs:
            raise IndicoInputError(f"One of {self.__options__} must be specified")
        self.update(and_(kwargs) if len(kwargs) > 1 else kwargs)


class SubmissionFilter(Filter):
    """
    Create a Filter when querying for WorkflowSubmissions.

    Args:
        input_filename (str): submissions with input file names containing this string
        status (str): submissions in this status. Options:
            [PROCESSING, PENDING_REVIEW, PENDING_ADMIN_REVIEW, COMPLETE, FAILED]
        retrieved(bool): Filter submissions on the retrieved flag
    Returns:
        dict containing query filter parameters
    """

    __options__ = ("input_filename", "status", "retrieved")

    def __init__(
        self, input_filename: str = None, status: str = None, retrieved: bool = None
    ):
        kwargs = {
            "input_filename": input_filename,
            "status": status.upper() if status else status,
            "retrieved": retrieved,
        }

        super().__init__(**kwargs)


class UserMetricsFilter(Filter):
    """
    Create a Filter when querying for UserSnapshots.

    Args:
        user_id (int): username to filter on
        user_email (str): email to filter for
    Returns:
        dict containing query filter parameters
    """
    __options__ = ("user_id", "user_email")

    def __init__(self, user_id: int = None, user_email: str = None):
        kwargs = {
            "userId": user_id,
            "userEmail": user_email
        }

        super().__init__(**kwargs)

class DocumentReportFilter(Filter):
    """
    Create a filter for the DocumentReport query.

    Args:
        workflow_id (int): workflow id
        submission_id (int): submission id
        status (str): submission status
        created_at_start_date (datetime): earliest creation date
        created_at_end_date (datetime): latest creation date
        updated_at_start_date (datetime): earliest update ddate
        updated_at_end_date (datetime): latest update date
    Returns:
        dict containing query filter parameters
    """
    __options__ = ("workflow_id", "submission_id", "status", "created_at_start_date", "created_at_end_date",
                   "updated_at_start_date", "updated_at_end_date")

    def __init__(self, submission_id: int = None, workflow_id: int = None, status: str = None, created_at_start_date: datetime = None,
                 created_at_end_date: datetime = None,
                 updated_at_start_date: datetime = None, updated_at_end_date: datetime = None
                 ):

        kwargs = {
            "workflowId": workflow_id,
            "id": submission_id,
            "status": status


        }
        if created_at_start_date is not None and created_at_end_date is not None:
            kwargs["createdAt"] = {
                "from": created_at_start_date.strftime('%Y-%m-%d') if created_at_start_date is not None else "",
                "to": created_at_end_date.strftime('%Y-%m-%d') if created_at_end_date is not None else "",
            }
        if updated_at_start_date is not None and updated_at_end_date is not None:
            kwargs["updatedAt"] = {
                "from": updated_at_start_date.strftime('%Y-%m-%d') if updated_at_start_date is not None else "",
                "to": updated_at_end_date.strftime('%Y-%m-%d') if updated_at_end_date is not None else "",
            }
        super().__init__(**kwargs)



from typing import List

from indico.types.base import BaseType


class DataFilePage(BaseType):
    """
    A DataFilePage in the Indico Platform

    DataFilePages give more specific info on individual datafile pages.

    Attributes:
        id (int): The DataFilePage id
        datafile_id (int): The Datafile id associated with this DataFilePage
        image (str): URL to where the image of the file is stored
        page_num (int): The page number of the Datafile associated with this info
        page_info (str): More meta information on the page, such as OCR stats
        thumbnail (str): URL to a smaller, thumbnail version of the DataFilePage image
        doc_start_offset (int): The index position where the page starts
        doc_end_offset (int): The index position where the page ends
    """

    id: int
    datafile_id: int
    image: str
    page_num: int
    page_info: str
    thumb_nail: str
    doc_start_offset: int
    doc_end_offset: int


class Datafile(BaseType):
    """
    A Datafile in the Indico Platform.

    Datafiles are primarlily used in retrieving stored results.

    Attributes:
        id (int): The Datafile id
        deleted (bool): Has the Datafile been marked for deletion
        name (str): Datafile name
        rainbow_url (str): URL of the Datafile within the Indico Platform
        status (str): Processing status of the Datafile
        status_meta (str): Contains metadata about the status of the Datafile, such as errors
        file_hash (str): Location of the Datafile
        file_size (int): Size of the file, in bytes
        file_type (str): Filetype of the Datafile
        page_ids (List[int]): A list of pagge ids associated with the Datafile
        num_pages (int): The number of pages in the Datafile
        pages (List[DataFilePage]): The DatafilePages associated with the Datafile
        failure_type (str): The type of failure associated with the Datafile, if an error was encountered
    """

    id: int
    deleted: bool
    name: str
    rainbow_url: str
    status: str
    status_meta: str
    file_hash: str
    file_size: int
    file_type: str
    page_ids: List[int]
    num_pages: int
    pages: List[DataFilePage]
    failure_type: str



"""
Setup for indico apis
"""
import versioneer
from setuptools import setup, find_packages

setup(
    name="indico-client",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    packages=find_packages(exclude=["tests"]),
    description="""A Python Wrapper for indico app API.""",
    license="MIT License (See LICENSE)",
    long_description=open("README.rst").read(),
    url="https://github.com/IndicoDataSolutions/indico-client-python",
    author="indico",
    author_email="engineering@indico.io",
    tests_require=["pytest>=5.2.1", "requests-mock>=1.7.0-7"],
    install_requires=[
        "msgpack==1.0.0",
        "msgpack-numpy==0.4.4.3",
        "numpy>=1.16.0",
        "Pillow>=6.2.0",
        "requests>=2.22.0",
        "setuptools>=41.4.0",
        "pandas>=1.0.3",
        'importlib-metadata ~= 1.0 ; python_version < "3.8"',
    ],
)



# -*- coding: utf-8 -*-

import json
import tempfile
from pathlib import Path
from typing import List

import pandas as pd
from indico.client.request import (
    Debouncer,
    GraphQLRequest,
    HTTPMethod,
    HTTPRequest,
    RequestChain,
)
from indico.errors import IndicoNotFound
from indico.queries.storage import UploadBatched, UploadImages
from indico.types.dataset import Dataset


class ListDatasets(GraphQLRequest):
    """
    List all of your datasets

    Options:
        limit (int, default=100): Max number of datasets to retrieve

    Returns:
        List[Dataset]

    Raises:

    """

    query = """
        query ListDatasets($limit: Int){
            datasetsPage(limit: $limit) {
                datasets {
                    id
                    name
                    rowCount
                }
            }
        }
    """

    def __init__(self, *, limit: int = 100):
        super().__init__(self.query, variables={"limit": limit})

    def process_response(self, response) -> Dataset:
        response = super().process_response(response)
        return [Dataset(**dataset) for dataset in response["datasetsPage"]["datasets"]]


class GetDataset(GraphQLRequest):
    """
    Retrieve a dataset description object

    Args:
        id (int): id of the dataset to query

    Returns:
        Dataset object

    Raises:

    """

    query = """
        query GetDataset($id: Int) {
            dataset(id: $id) {
                id
                name
                rowCount
                status
                permissions
                datacolumns {
                    id
                    name
                }
                labelsets{
                    id
                    name
                }
            }
        }
    """

    def __init__(self, id: int):
        super().__init__(self.query, variables={"id": id})

    def process_response(self, response) -> Dataset:
        response = super().process_response(response)
        if "dataset" not in response or not isinstance(response["dataset"], dict):
            raise IndicoNotFound("Failed to find dataset")
        return Dataset(**response["dataset"])


class GetDatasetFileStatus(GetDataset):
    """
    Get the status of dataset file upload

    Args:
        id (int): id of the dataset to query

    Returns:
        status (str): DOWNLOADED or FAILED

    Raises:

    """

    query = """
        query DatasetUploadStatus($id: Int!) {
            dataset(id: $id) {
                id
                status
                files {
                    id
                    name
                    deleted
                    fileSize
                    rainbowUrl
                    fileType
                    fileHash
                    status
                    statusMeta
                    failureType
                }
            }
        }
    """


class GetDatasetStatus(GraphQLRequest):
    """
    Get the status of a dataset

    Args:
        id (int): id of the dataset to query

    Returns:
        status (str): COMPLETE or FAILED

    Raises:

    """

    query = """
        query datasetStatus($id: Int!) {
            dataset(id: $id) {
                status
            }
        }
    """

    def __init__(self, id: int):
        super().__init__(self.query, variables={"id": id})

    def process_response(self, response) -> str:
        return response["data"]["dataset"]["status"]


class CreateDataset(RequestChain):
    """
    Create a dataset and upload the associated files.
    Args:
        name (str): Name of the dataset
        files (List[str]): List of pathnames to the dataset files

    Options:
        dataset_type (str): Type of dataset to create [TEXT, DOCUMENT, IMAGE]
        wait (bool, default=True): Wait for the dataset to upload and finish
    Returns:
        Dataset object
    Raises:
        IndicoError
    """

    previous = None

    def __init__(
        self,
        name: str,
        files: List[str],
        wait: bool = True,
        dataset_type: str = "TEXT",
        from_local_images: bool = False,
        image_filename_col: str = "filename",
        batch_size: int = 20,
    ):
        self.files = files
        self.name = name
        self.wait = wait
        self.dataset_type = dataset_type
        self.from_local_images = from_local_images
        self.image_filename_col = image_filename_col
        self.batch_size = batch_size
        super().__init__()

    def requests(self):
        if self.from_local_images:
            self.dataset_type = "IMAGE"
            # Assume image filenames are in the same directory as the csv with
            # image labels and that there is a column representing their name
            df = pd.read_csv(self.files)
            img_filenames = df[self.image_filename_col].tolist()
            img_filepaths = [
                str(Path(self.files).parent / imgfn) for imgfn in img_filenames
            ]
            yield UploadBatched(
                img_filepaths,
                batch_size=self.batch_size,
                request_cls=UploadImages,
            )
            df["urls"] = self.previous
            with tempfile.TemporaryDirectory() as tmpdir:
                image_csv_path = str(Path(tmpdir) / "image_urls.csv")
                df.to_csv(image_csv_path)
                yield _UploadDatasetFiles(files=[image_csv_path])
        else:
            yield UploadBatched(
                files=self.files,
                batch_size=self.batch_size,
                request_cls=_UploadDatasetFiles,
            )
        file_metadata = self.previous
        yield CreateEmptyDataset(name=self.name, dataset_type=self.dataset_type)
        yield _AddFiles(dataset_id=self.previous.id, metadata=file_metadata)
        dataset_id = self.previous.id
        yield GetDatasetFileStatus(id=dataset_id)
        debouncer = Debouncer()
        while not all(
            f.status in ["DOWNLOADED", "FAILED"] for f in self.previous.files
        ):
            yield GetDatasetFileStatus(id=self.previous.id)
            debouncer.backoff()
        csv_files = [f.id for f in self.previous.files if f.file_type == "CSV"]
        non_csv_files = [f.id for f in self.previous.files if f.file_type != "CSV"]

        if csv_files:
            yield _ProcessCSV(dataset_id=dataset_id, datafile_ids=csv_files)
        elif non_csv_files:
            yield _ProcessFiles(dataset_id=dataset_id, datafile_ids=non_csv_files)
        yield GetDatasetFileStatus(id=dataset_id)
        debouncer = Debouncer()
        if self.wait is True:
            while not all(
                [f.status in ["PROCESSED", "FAILED"] for f in self.previous.files]
            ):
                yield GetDatasetFileStatus(id=dataset_id)
                debouncer.backoff()
        yield GetDataset(id=dataset_id)


class _UploadDatasetFiles(HTTPRequest):
    def __init__(self, files: List[str]):
        super().__init__(
            method=HTTPMethod.POST, path="/storage/files/upload", files=files
        )


class DeleteDataset(GraphQLRequest):
    """
    Delete a dataset

    Args:
        id (int): ID of the dataset

    Returns:
        success (bool): The success of the operation

    Raises:

    """

    query = """
    mutation deleteDataset($id: Int!) {
        deleteDataset(id: $id) {
            success
        }
    }
    """

    def __init__(self, id):
        super().__init__(self.query, variables={"id": id})

    def process_response(self, response):
        return super().process_response(response)["deleteDataset"]["success"]


class CreateEmptyDataset(GraphQLRequest):
    query = """
    mutation($name: String!, $datasetType: DatasetType) {
        createDataset(name: $name, datasetType: $datasetType) {
            id
            name
        }
    }
    """

    def __init__(self, name: str, dataset_type: str = None):
        if not dataset_type:
            dataset_type = "TEXT"

        super().__init__(
            self.query, variables={"name": name, "datasetType": dataset_type}
        )

    def process_response(self, response):
        return Dataset(**super().process_response(response)["createDataset"])


class _AddFiles(GraphQLRequest):
    query = """
    mutation AddFiles($datasetId: Int!, $metadata: JSONString!){
        addDatasetFiles(datasetId: $datasetId, metadataList: $metadata) {
            id
            status
        }
    }
    """
    

    def __init__(self, dataset_id: int, metadata: List[str]):
        super().__init__(
            self.query,
            variables={"datasetId": dataset_id, "metadata": json.dumps(metadata)},
        )

    def process_response(self, response):
        return Dataset(**super().process_response(response)["addDatasetFiles"])


class AddFiles(RequestChain):
    """
    Add files to a dataset

    Args:
        dataset_id (int): ID of the dataset
        files (List[str]): List of pathnames to the dataset files

    Options:
        wait (bool, default=True): Block while polling for status of files
        batch_size (int, default=20): Batch size for uploading files

    Returns:
        Dataset

    Raises:

    """

    previous = None

    def __init__(
        self,
        dataset_id: int,
        files: List[str],
        wait: bool = True,
        batch_size: int = 20,
    ):
        self.dataset_id = dataset_id
        self.files = files
        self.wait = wait
        self.batch_size = batch_size
        super().__init__()

    def requests(self):
        yield UploadBatched(
            files=self.files,
            batch_size=self.batch_size,
            request_cls=_UploadDatasetFiles,
        )
        yield _AddFiles(dataset_id=self.dataset_id, metadata=self.previous)
        yield GetDatasetFileStatus(id=self.dataset_id)
        debouncer = Debouncer()
        while not all(
            f.status in ["DOWNLOADED", "FAILED", "PROCESSED"]
            for f in self.previous.files
        ):
            yield GetDatasetFileStatus(id=self.previous.id)
            debouncer.backoff()


class _ProcessFiles(GraphQLRequest):
    query = """
    mutation (
        $datasetId: Int!,
        $datafileIds: [Int]) {
        addDataFiles(
          datasetId: $datasetId,
          datafileIds: $datafileIds) {
            id
            name
        }
    }
    """

    def __init__(self, dataset_id: int, datafile_ids: List[int]):
        super().__init__(
            self.query,
            variables={"datasetId": dataset_id, "datafileIds": datafile_ids},
        )

    def process_response(self, response):
        return Dataset(**super().process_response(response)["addDataFiles"])


class _ProcessCSV(GraphQLRequest):
    query = """
    mutation ($datasetId: Int!, $datafileIds: [Int]) {
        addDataCsv(datasetId: $datasetId, datafileIds: $datafileIds) {
            id
            name
        }
    }
    """

    def __init__(self, dataset_id: int, datafile_ids: List[int]):
        super().__init__(
            self.query, variables={"datasetId": dataset_id, "datafileIds": datafile_ids}
        )

    def process_response(self, response):
        return Dataset(**super().process_response(response)["addDataCsv"])


class ProcessFiles(RequestChain):
    """
    Process files associated with a dataset and add corresponding data to the dataset

    Args:
        dataset_id (int): ID of the dataset
        datafile_ids (List[str]): IDs of the datafiles to process
        wait (bool): Block while polling for status of files


    Returns:
        Dataset

    Raises:

    """

    def __init__(
        self,
        dataset_id: int,
        datafile_ids: List[int],
        wait: bool = True,
    ):
        self.dataset_id = dataset_id
        self.datafile_ids = datafile_ids
        self.wait = wait

    def requests(self):
        yield _ProcessFiles(self.dataset_id, self.datafile_ids)
        debouncer = Debouncer()
        yield GetDatasetFileStatus(id=self.dataset_id)
        if self.wait:
            while not all(
                f.status in ["PROCESSED", "FAILED"] for f in self.previous.files
            ):
                yield GetDatasetFileStatus(id=self.dataset_id)
                debouncer.backoff()


class ProcessCSV(RequestChain):
    """
    Process CSV associated with a dataset and add corresponding data to the dataset

    Args:
        dataset_id (int): ID of the dataset
        datafile_ids (List[str]): IDs of the CSV datafiles to process
    Options:
        wait (bool, default=True): Block while polling for status of files

    Returns:
        Dataset

    Raises:

    """

    def __init__(self, dataset_id: int, datafile_ids: List[int], wait: bool = True):
        self.dataset_id = dataset_id
        self.datafile_ids = datafile_ids
        self.wait = wait

    def requests(self):
        yield _ProcessCSV(self.dataset_id, self.datafile_ids)
        debouncer = Debouncer()
        yield GetDatasetFileStatus(id=self.dataset_id)
        if self.wait:
            while not all(
                f.status in ["PROCESSED", "FAILED"] for f in self.previous.files
            ):
                yield GetDatasetFileStatus(id=self.dataset_id)
                debouncer.backoff()



from typing import List

from indico.types.base import BaseType
from indico.types.datafile import Datafile


class DataColumn(BaseType):
    """
    A DataColumn in the Indico Platform

    Attributes:
        id (int): The DataColumn id
        name (str): DataColumn name
    """

    id: int
    name: str


class LabelSet(BaseType):
    """
    A LabelSet in the Indico Platform

    Attributes:
        id (int): The LabelSet id
        name (str): LabelSet name
    """

    id: int
    name: str


class Dataset(BaseType):
    """
    A Dataset in the Indico Platform

    Attributes:
        id (int): The Dataset id
        name (str): Dataset name
        row_count (int): Number of rows in the dataset
        permissions (str): Permissions on the dataset
        files (List[Datafile]): Names of the file(s) included in the dataset
        labelsets (List[LabelSet]): LabelSets associated with this dataset
        datacolumns (List[DataColumn]): DataColumn(s) of the dataset
    """

    id: int
    name: str
    row_count: int
    status: str
    permissions: str
    files: List[Datafile]
    labelsets: List[LabelSet]
    datacolumns: List[DataColumn]

    def labelset_by_name(self, name: str) -> LabelSet:
        return next(l for l in self.labelsets if l.name == name)

    def datacolumn_by_name(self, name: str) -> DataColumn:
        return next(l for l in self.datacolumns if l.name == name)



from datetime import datetime
from indico.errors import IndicoInputError
from indico.client.request import GraphQLRequest
from indico.types import BaseType
from indico.types.workflow_metrics import WorkflowMetricsOptions, WorkflowMetrics
from typing import List


class _WorkflowMetric(BaseType):
    metrics: WorkflowMetrics


class _TopWorkflowMetric(BaseType):
    workflows: List[_WorkflowMetric]


class GetWorkflowMetrics(GraphQLRequest):
    """
    Requests detailed workflow metric data.
    Includes daily and total submission counts,
    Review queue counts, and Straight through processing details.
    Query can be configured to include only specific metrics
    by passing in one of WorkflowOptions for SUBMISSIONS,
    REVIEW, or STRAIGHT_THROUGH_PROCESSING.

    Args:
        options (List[WorkflowMetricsOptions]): specific metrics to return.
        start_date (datetime): start date for metrics data.
        end_date (datetime): end date for metrics data. defaults to today.
        workflow_ids (List[int]): ids of specific workflows to query.

    """
    __MAP_WORKFLOW_KEYS = {
        WorkflowMetricsOptions.SUBMISSIONS: """
        firstSubmittedDate
               submissions {               
                  aggregate {
                    submitted
                    completed
                    completedInReview
                    completedReviewQueue
                    completedExceptionQueue
                    rejectedInReview
                  }
                  daily {
                    date
                    submitted
                    completed
                    completedInReview
                    completedReviewQueue
                    completedExceptionQueue
                    rejectedInReview
                  }
               }""",
        WorkflowMetricsOptions.REVIEW: """ queues {
                  dailyCumulative {
                    date
                    subsOnQueue
                    hoursOnQueue
                    avgAgeInQueue # (num_hours / num_subs )
                  }
               }
               """,
        WorkflowMetricsOptions.STRAIGHT_THROUGH_PROCESSING: """
              straightThroughProcessing {
                workflow {
                  daily {
                    date
                    reviewNumerator
                    autoReviewNumerator
                    reviewDenom
                    autoReviewDenom
                    reviewStpPct
                    autoReviewStpPct
                  }
                }
                model {
                  modelGroupId
                  name
                  aggregate {
                    reviewNumerator
                    autoReviewNumerator
                    reviewDenom
                    autoReviewDenom
                    reviewStpPct
                    autoReviewStpPct
                  }
                  daily {
                    date
                    reviewNumerator
                    autoReviewNumerator
                    reviewDenom
                    autoReviewDenom
                    reviewStpPct
                    autoReviewStpPct
                  }
                  classMetrics {
                    className
                    aggregate {
                      reviewNumerator
                      autoReviewNumerator
                      reviewDenom
                      autoReviewDenom
                      reviewStpPct
                      autoReviewStpPct                    
                    }
                    daily {
                      date
                      reviewNumerator
                      autoReviewNumerator
                      reviewDenom
                      autoReviewDenom
                      reviewStpPct
                      autoReviewStpPct
                    }
                  }
                }
            }
                """,
        WorkflowMetricsOptions.TIME_ON_TASK: """
               timeOnTask {
                  aggregate { 
                    avgMinsPerDoc
                    avgMinsPerDocReview
                    avgMinsPerDocExceptions
                  }
                  daily {
                    date
                    avgMinsPerDoc
                    avgMinsPerDocReview
                    avgMinsPerDocExceptions
                  }
               }
        """

    }
    query = """
query ($workflowIds: [Int]!, $startDate: Date, $endDate:Date) {
    workflows(workflowIds: $workflowIds){
        workflows {
            metrics(startDate:$startDate, endDate:$endDate){
               firstSubmittedDate
               workflowId
               __QUERY_OPTS__
              }
            }
        }
}
"""

    def __init__(self, options: List[WorkflowMetricsOptions], start_date: datetime, end_date: datetime,
                 workflow_ids: List[int]):
        self.query = self.__map_query_values(options)
        if workflow_ids is None or start_date is None:
            raise IndicoInputError("Must specify date and workflow id")
        if end_date is None:
            end_date = datetime.now()
        super().__init__(self.query, variables={"startDate": start_date.strftime('%Y-%m-%d'),
                                                "endDate": end_date.strftime('%Y-%m-%d'), "workflowIds": workflow_ids})

    def process_response(self, response) -> List[WorkflowMetrics]:
        list_of_metrics = _TopWorkflowMetric(**super().process_response(response)["workflows"]).workflows
        return list(map(lambda x: x.metrics, list_of_metrics))

    def __map_query_values(self, options: List[WorkflowMetricsOptions]):
        daily = ' '
        if len(options) < 1:
            daily = ' '.join([self.__MAP_WORKFLOW_KEYS[a] for a in self.__MAP_WORKFLOW_KEYS.keys()])
        else:
            daily = ' '.join([self.__MAP_WORKFLOW_KEYS[a] for a in options])
        query = self.query.replace("__QUERY_OPTS__", daily)
        return query



from indico.types.base import BaseType
from indico.types.model import Model


class ModelGroup(BaseType):
    """
    A Model Group in the Indico Platform.

    Think of Model Groups as a container for individual models that are trained each
    time you click the Retrain button in Review or label more data in Teach. One of these
    models will be the current, "selected" model. It's usually the latest model but can also
    be the best performing.

    Attributes:
        id (int): The model group id
        name (str): The name of the model group
        status (str): "CREATED", "TRAINING", "COMPLETE" or "FAILED"
        selected model (Model): A Model object representing the selected (active) model.
        task_type (str): Type of machine learning task the model solves
    """

    id: int
    name: str
    status: str
    selected_model: Model
    task_type: str
